"""
autowire.helpers
================

Helpers.

"""
import contextlib
import functools

from autowire.base import BaseContext, BaseResource

from .builtins import this
from .utils import RefCounter


def autowired(fn, *dependencies: BaseResource):
    """
    Convert plain contextmanager to resource implementation by autowiring
    dependencies. ::

        foo = Resource('foo', __name__)
        bar = Resource('bar', __name__)

        @contextlib.contextmanager
        def with_foo(bar):
            value = baz(bar)
            try:
                yield value
            finally:
                value.teardown()

        impl = autowired(with_foo, bar)  # Convert to implementation.
        foo.impl(impl)

    """
    @contextlib.contextmanager
    def implementation(context: BaseContext):
        # Resolve dependencies using context.
        def with_dependencies(fn, dependencies):
            if dependencies:
                first = dependencies[0]
                rest = dependencies[1:]
                with context.resolve(first) as value:
                    fn = functools.partial(fn, value)
                    yield from with_dependencies(fn, rest)
            else:
                with fn() as value:
                    yield value
        # Call implemenation with autowired resources.
        yield from with_dependencies(fn, dependencies)
    return implementation


def shared(impl):
    """
    Convert implementation to shared resource implementation.

    Shared resource shares contexxt using reference counting.
    When you try to resolve shared resource that already resolve but not teared
    down, It brings previously resolved resouce.

    The resource will be tied into resource's providing context.
    It means that children's context will not be able to be used
    in this resource. ::

        res1 = Resource('res1', __name__)
        res2 = Resource('res2', __name__)

        @res2.impl
        @shared
        @contextlib.contextmanager
        def with_res1(context):
            with context.resolve(res2) as res1_factory:
                res1 = res1_factory.create_res1()
                try:
                    yield res1
                finally:
                    res1.teardown()

        context = Context()

        child = Context(context)

        @child.autowired(res2)
        def with_res2():
            yield ...

        # Get ResourceNotProvidedError, even though child provides res2.
        # Because res1 was shared in parent context.
        with child.resolve(res1):
            ...

    """

    counter = None

    @functools.wraps(impl)
    @contextlib.contextmanager
    def wrapper(context: BaseContext):
        nonlocal counter
        with context.resolve(this) as resource:
            if counter is None:
                provider = context.provided_by(resource)  # Find me
                counter = RefCounter(impl(provider))
            try:
                with counter as value:
                    yield value
            finally:
                if counter.count == 0:
                    counter = None
    return wrapper



"""
autowire.context.impl
=====================

Context's implementations

"""
import itertools

from autowire.base import BaseContext, BaseResource
from autowire.decorators import autowired
from autowire.utils import apply_decorators, as_contextmanager

from .root import root_context


class Context(BaseContext):
    """
    Context implementation.
    """

    def __init__(self, parent: BaseContext=root_context):
        self.parent = parent
        self.implementations = {}

    def get_resource_impl(self, resource: BaseResource):
        """Get resource implementation from this context."""
        return self.implementations[resource.canonical_name]

    def provided_by(self, resource: BaseResource):
        """Find context that provides resource."""
        if resource.canonical_name in self.implementations:
            return self
        elif self.parent is not None:
            return self.parent.provided_by(resource)

    def set_resource_impl(self, resource: BaseResource, impl):
        """Set resource implementation to this context."""
        self.implementations[resource.canonical_name] = impl

    def provide(self, resource: BaseResource):
        """
        Provide implementation for resource in this context. ::

            @context.provide_autowired(resource, dependency1, dependency2)
            @contextlib.contextmanager
            def create_resource(dependency1, dependency2):
                yield dependency1.make_resource(dependency2)

        """
        def decorator(implementation):
            self.set_resource_impl(resource, implementation)
            return implementation
        return decorator

    def provide_with_decorators(self, resource, *decorators):
        def decorator(fn):
            decorated = apply_decorators(fn, decorators)
            self.provide(resource)(decorated)
            return fn
        return decorator

    def provide_from_func(self, resource: BaseResource,
                          *dependencies: BaseResource,
                          decorators=()):
        """
        Provide resource from function implementation. ::

            @context.provide_from_func(resource, dependency)
            def get_resource(dependency):
                return os.path.join(dependency, 'resource.json')

        """

        return self.provide_with_decorators(
            resource,
            as_contextmanager,
            autowired(*dependencies),
            *decorators
        )

    def provide_autowired(self, resource: BaseResource,
                          *dependencies: BaseResource,
                          decorators=()):
        """
        Provide implementation with autowired dependencies ::

            @context.provide_autowired(resource, dependency1, dependency2)
            @contextlib.contextmanager
            def create_resource(dependency1, dependency2):
                yield dependency1.make_resource(dependency2)

        """
        return self.provide_with_decorators(
            resource,
            autowired(*dependencies),
            *decorators
        )



"""
autowire.decorators
===================

Decorators for autowire.

"""
from . import helpers
from .base import BaseResource


def autowired(*dependencies: BaseResource):
    """
    Convert a plain function to resource implementation with dependencies.
    It changes interface of decorated function to implementation interface.
    """
    def decorator(fn):
        return helpers.autowired(fn, *dependencies)
    return decorator


# It already has decorator interface
shared = helpers.shared



"""
autowire.resource.impl
======================

Resource implementations.

"""
import contextlib

from autowire.base import BaseContext, BaseResource
from autowire.exc import ResourceNotProvidedError
from autowire.decorators import autowired
from autowire.utils import apply_decorators, as_contextmanager


def create_default_impl(name: str, namespace: str):
    @contextlib.contextmanager
    def default_impl(context: BaseContext):
        """No such resource."""
        raise ResourceNotProvidedError(
            "No such resource {namespace}.{name}".format(
                namespace=namespace,
                name=name,
            ))
        yield None
    return default_impl


class Resource(BaseResource):
    """
    Resource class of some functionalities like default implementation.

    """

    def __init__(self, name, namespace):
        super().__init__(name, namespace)
        self.default_implementation = create_default_impl(name, namespace)

    @property
    def default_implementation(self):
        return self._default_implementation

    @default_implementation.setter
    def default_implementation(self, default_implementation):
        self._default_implementation = default_implementation

    def impl(self, implementation, decorators=()):
        """Set default implementation of resource."""
        self.default_implementation = implementation
        return implementation

    def with_decorators(self, *decorators):
        """
        Set default implementation with decorators.
        It sets implementation with decorators but returns original function
        not decorated function so that does not change function's interface.

        """
        def decorator(fn):
            decorated = apply_decorators(fn, decorators)
            self.impl(decorated)
            return fn
        return decorator

    def autowired(self, *dependencies: BaseResource, decorators=()):
        """
        Set default implementation to resource. ::

            @builder.autowired(dependency1, dependency2)
            @contextlib.contextmanager
            def create_resource(dependency1, dependency2):
                yield dependency1.make_resource(dependency2)

        To apply decorators to implementation ::

            @resource.autowired(dependency, decorators=[decorator])
            @contextlib.contextmanager
            def create_resource(dependency):
                return dependency.make_resource()

        """
        return self.with_decorators(autowired(*dependencies), *decorators)

    def from_func(self, *dependencies, decorators=()):
        """
        Get resource from function implementation. ::

            @resource.from_func(dependency)
            def get_resource(dependency):
                return os.path.join(dependency, 'resource.json')

        """
        return self.with_decorators(
            as_contextmanager,
            autowired(*dependencies),
            *decorators
        )



"""
autowire.base
=============

Base definitions of autowire.

"""
import abc
import sys

MAJOR, MINOR = sys.version_info[:2]
PY33 = MAJOR == 3 and MINOR >= 3


# abc.abstractproperty was deprecated since Python 3.3
if PY33:
    def abstractproperty(getter):
        return property(abc.abstractmethod(getter))
else:
    abstractproperty = abc.abstractproperty


class BaseResource(object, metaclass=abc.ABCMeta):
    """
    Decalarative contextual resource definition.
    """
    def __init__(self, name: str, namespace: str) -> None:
        """
        Create a resource with name.
        Namespace generally be the module's name.

        `name` cannot include any dot(.) characters.

            >>> resource = Resource('name', __name__)

        """
        super().__init__()
        self.name = name
        self.namespace = namespace

    def is_valid_name(self, name: str) -> bool:
        """
        Resource name validation.
        """
        return '.' not in name

    @property
    def canonical_name(self) -> str:
        """
        Canonical name of resource.

        It's <namespace>.<name>

        """
        return self.namespace + '.' + self.name

    @abstractproperty
    def default_implementation(self):
        pass

    def __repr__(self):
        return "{cls}({name!r}, {namespace!r})".format(
            cls=type(self).__name__,
            name=self.name,
            namespace=self.namespace
        )


class BaseContext(object, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def get_resource_impl(self, resource: BaseResource):
        """Get resource implementation from this context."""
        pass

    @abc.abstractmethod
    def provided_by(self, resource: BaseResource) -> 'BaseContext':
        """Find context that provides resource."""
        pass

    def resolve(self, resource: BaseResource):
        """Resolve resource in this context."""
        impl = self.find_resource_impl(resource)
        return impl(self)

    def find_resource_impl(self, resource: BaseResource):
        """Find resource implementation from this context and its parents."""
        context = self.provided_by(resource)
        return context.get_resource_impl(resource)
